@@ -1,16 +1,19 @@
 import json
 import os
+import signal
 import tempfile
 import logging
 
 from twisted.internet.protocol import ProcessProtocol
-from twisted.internet import defer
+from twisted.internet import defer, task
 from twisted.protocols.basic import LineReceiver
 
 
 def catchall_errback(logger):
     def eb(failure):
-        logger.exception(failure.value)
+        logger.error(
+            '%s: %s\n%s', failure.type.__name__, failure.value,
+            failure.getTraceback())
     return eb
 
 
@@ -67,24 +70,84 @@ class Job(object):
         self.board_name = board_name
         self.reactor = reactor
         self._json_file = None
+        self._source_lock = defer.DeferredLock()
+        self._checkCancel_call = task.LoopingCall(self._checkCancel)
+
+    def _checkCancel(self):
+        return self._source_lock.run(
+            self.source.jobCheckForCancellation, self.board_name).addCallback(
+            self._maybeCancel)
+
+    def _maybeCancel(self, cancel):
+        if cancel:
+            self._protocol.transport.signalProcess(signal.SIGINT)
 
     def run(self):
+        d = self.source.getLogFileForJobOnBoard(self.board_name)
+        return d.addCallback(self._run).addErrback(
+            catchall_errback(self.logger))
+
+    def _run(self, log_file):
         d = defer.Deferred()
-        json_data, log_file = self.job_data
+        json_data = self.job_data
         fd, self._json_file = tempfile.mkstemp()
         with os.fdopen(fd, 'wb') as f:
             json.dump(json_data, f)
+        self._protocol = DispatcherProcessProtocol(
+            d, log_file, self.source, self.board_name)
         self.reactor.spawnProcess(
-            DispatcherProcessProtocol(
-                d, log_file, self.source, self.board_name),
-            self.dispatcher, args=[
+            self._protocol, self.dispatcher, args=[
                 self.dispatcher, self._json_file, '--oob-fd', '3'],
             childFDs={0:0, 1:'r', 2:'r', 3:'r'}, env=None)
+        self._checkCancel_call.start(10)
+        d.addBoth(self._exited)
+        return d
+
+    def _exited(self, result):
+        self.logger.info("job finished on %s", self.job_data['target'])
+        if self._json_file is not None:
+            os.unlink(self._json_file)
+        self.logger.info("reporting job completed")
+        self._source_lock.run(self._checkCancel_call.stop)
+        return self._source_lock.run(
+            self.source.jobCompleted, self.board_name).addCallback(
+            lambda r:result)
+
+
+class SimplePP(ProcessProtocol):
+    def __init__(self, d):
+        self.d = d
+    def processEnded(self, reason):
+        self.d.callback(None)
+
+
+class MonitorJob(object):
+
+    logger = logging.getLogger(__name__ + '.MonitorJob')
+
+    def __init__(self, job_data, dispatcher, source, board_name, reactor):
+        self.job_data = job_data
+        self.dispatcher = dispatcher
+        self.source = source
+        self.board_name = board_name
+        self.reactor = reactor
+        self._json_file = None
+
+    def run(self):
+        d = defer.Deferred()
+        json_data = self.job_data
+        fd, self._json_file = tempfile.mkstemp()
+        with os.fdopen(fd, 'wb') as f:
+            json.dump(json_data, f)
+        self.reactor.spawnProcess(
+            SimplePP(d), 'setsid', childFDs={0:0, 1:1, 2:2},
+            env=None, args=[
+                'setsid', 'lava-scheduler-monitor', self.dispatcher,
+                str(self.board_name), self._json_file])
         d.addBoth(self._exited)
         return d
 
     def _exited(self, result):
-        self.logger.info("job finished on %s", self.job_data[0]['target'])
         if self._json_file is not None:
             os.unlink(self._json_file)
         return result
@@ -138,7 +201,7 @@ class Board(object):
     not always do what you expect.  So don't mess around in that way please.
     """
 
-    job_cls = Job
+    job_cls = MonitorJob
 
     def __init__(self, source, board_name, dispatcher, reactor, job_cls=None):
         self.source = source
@@ -199,7 +262,9 @@ class Board(object):
             self._maybeStartJob, self._ebCheckForJob)
 
     def _ebCheckForJob(self, result):
-        self.logger.exception(result.value)
+        self.logger.error(
+            '%s: %s\n%s', result.type.__name__, result.value,
+            result.getTraceback())
         self._maybeStartJob(None)
 
     def _finish_stop(self):
@@ -227,16 +292,11 @@ class Board(object):
         d = self.running_job.run()
         d.addCallbacks(self._cbJobFinished, self._ebJobFinished)
 
-    def _cbJobFinished(self, result):
-        self.logger.info("reporting job completed")
-        self.source.jobCompleted(
-            self.board_name).addCallback(self._cbJobCompleted)
-
     def _ebJobFinished(self, result):
         self.logger.exception(result.value)
         self._checkForJob()
 
-    def _cbJobCompleted(self, result):
+    def _cbJobFinished(self, result):
         self.running_job = None
         if self._stopping_deferreds:
             self._finish_stop()
